@@ -198,6 +198,72 @@ class ZopePageTemplatesTest(RenderTestCase):
         else:
             self.fail("Expected exception")
 
+    def test_exists_error_leak(self):
+        body = '''\
+        <?xml version="1.0"?>
+        <root>
+        <one tal:condition="exists: var_does_not_exists" />
+        <two tal:attributes="my_attr dict()['key-does-not-exist']" />
+        </root>'''
+        template = self.from_string(body, strict=False)
+        try:
+            template()
+        except RenderError:
+            exc = sys.exc_info()[1]
+            self.assertNotIn('var_does_not_exists', str(exc))
+        else:
+            self.fail("Expected exception")
+
+    def test_sys_exc_info_is_clear_after_pipe(self):
+        body = """<div tal:content="y|nothing"></div><span tal:content="error()" />"""
+        template = self.from_string(body, strict=False)
+
+        got = template.render(error=sys.exc_info)
+        self.assertTrue('<span>(None, None, None)</span>' in got)
+
+    def test_render_macro_include_subtemplate_containing_error(self):
+        macro_inner = self.from_string('''<two tal:attributes="my_attr dict()['key-does-not-exist']" />''')
+        macro_wrap = self.from_string('''<one tal:content="macro_inner()" />''')
+        template = self.from_string(
+            '''
+            <tal:defs define="translate string:">
+              <span i18n:translate="">foo</span>
+              <metal:macro use-macro="macro" />
+            </tal:defs>
+            ''')
+        try:
+            result = template(macro=macro_wrap, macro_inner=macro_inner)
+        except RenderError:
+            exc = sys.exc_info()[1]
+            self.assertTrue(isinstance(exc, KeyError))
+            self.assertIn(''''key-does-not-exist'
+
+ - Expression: "dict()['key-does-not-exist']"
+ - Filename:   <string>
+ - Location:   (line 1: col 29)
+ - Expression: "macro_inner()"
+ - Filename:   <string>
+ - Location:   (line 1: col 18)
+ - Expression: "macro"
+ - Filename:   <string>
+ - Location:   (line 4: col 38)
+''', str(exc))
+        else:
+            self.fail("Expected exception")
+
+    def test_render_error_macro_include(self):
+        body = """<metal:block use-macro='"bad"' />"""
+        template = self.from_string(body, strict=False)
+
+        try:
+            template()
+        except RenderError:
+            exc = sys.exc_info()[1]
+            self.assertTrue(isinstance(exc, AttributeError))
+            self.assertIn('bad', str(exc))
+        else:
+            self.fail("Expected exception")
+
     @error("""<tal:dummy attributes=\"dummy 'dummy'\" />""")
     def test_attributes_on_tal_tag_fails(self, body, exc):
         self.assertTrue(body[exc.offset:].startswith('dummy'))
@@ -319,6 +385,17 @@ class ZopePageTemplatesTest(RenderTestCase):
         rendered = template(test=object())
         self.assertTrue('object' in rendered)
 
+    def test_on_error_handler(self):
+        exc = []
+        handler = exc.append
+        template = self.from_string(
+            '<tal:block on-error="string:error">${test}</tal:block>',
+            on_error_handler=handler
+        )
+        rendered = template()
+        self.assertEquals(len(exc), 1)
+        self.assertEquals(exc[0].__class__.__name__, "NameError")
+
     def test_object_substitution_coerce_to_str(self):
         template = self.from_string('${test}', translate=None)
 
@@ -632,7 +709,7 @@ class ZopeTemplatesTestSuite(RenderTestCase):
             # source when debugging
             self.shortDescription = lambda: input_path
 
-            # When input path contaiins the string 'implicit-i18n', we
+            # When input path contains the string 'implicit-i18n', we
             # enable "implicit translation".
             implicit_i18n = 'implicit-i18n' in input_path
             implicit_i18n_attrs = ("alt", "title") if implicit_i18n else ()
@@ -689,7 +766,9 @@ class ZopeTemplatesTestSuite(RenderTestCase):
                 content_type, encoding = detect_encoding(
                     output, template.default_encoding)
 
-            want = output.decode(encoding)
+            # Newline normalization across platforms
+            want = '\n'.join(output.decode(encoding).splitlines())
+            got = '\n'.join(got.splitlines())
 
             if checker.check_output(want, got, 0) is False:
                 from doctest import Example
